@@ -33,7 +33,7 @@ static void add_pad_group_from_libinput(struct wlr_tablet_pad *pad,
 			++group->ring_count;
 		}
 	}
-	group->rings = calloc(group->ring_count, sizeof(unsigned int));
+	group->rings = calloc(sizeof(unsigned int), group->ring_count);
 	if (group->rings == NULL) {
 		goto group_fail;
 	}
@@ -50,7 +50,7 @@ static void add_pad_group_from_libinput(struct wlr_tablet_pad *pad,
 			++group->strip_count;
 		}
 	}
-	group->strips = calloc(group->strip_count, sizeof(unsigned int));
+	group->strips = calloc(sizeof(unsigned int), group->strip_count);
 	if (group->strips == NULL) {
 		goto group_fail;
 	}
@@ -66,7 +66,7 @@ static void add_pad_group_from_libinput(struct wlr_tablet_pad *pad,
 			++group->button_count;
 		}
 	}
-	group->buttons = calloc(group->button_count, sizeof(unsigned int));
+	group->buttons = calloc(sizeof(unsigned int), group->button_count);
 	if (group->buttons == NULL) {
 		goto group_fail;
 	}
