@@ -46,6 +46,7 @@ struct touch_state {
 	struct wl_listener down;
 	struct wl_listener up;
 	struct wl_listener motion;
+	struct wl_listener cancel;
 	struct wl_list link;
 	void *data;
 };
@@ -75,28 +76,35 @@ static void output_frame_notify(struct wl_listener *listener, void *data) {
 	int32_t width, height;
 	wlr_output_effective_resolution(wlr_output, &width, &height);
 
-	wlr_output_attach_render(wlr_output, NULL);
-	wlr_renderer_begin(sample->renderer, wlr_output->width, wlr_output->height);
-	wlr_renderer_clear(sample->renderer, (float[]){0.25f, 0.25f, 0.25f, 1});
+	struct wlr_output_state output_state;
+	wlr_output_state_init(&output_state);
+	struct wlr_render_pass *pass = wlr_output_begin_render_pass(wlr_output, &output_state, NULL, NULL);
+	wlr_render_pass_add_rect(pass, &(struct wlr_render_rect_options){
+		.box = { .width = width, .height = height },
+		.color = { 0.25, 0.25, 0.25, 1 },
+	});
 
 	struct touch_point *p;
 	wl_list_for_each(p, &sample->touch_points, link) {
 		int x = (int)(p->x * width) - sample->cat_texture->width / 2;
 		int y = (int)(p->y * height) - sample->cat_texture->height / 2;
-		wlr_render_texture(sample->renderer, sample->cat_texture,
-			wlr_output->transform_matrix, x, y, 1.0f);
+		wlr_render_pass_add_texture(pass, &(struct wlr_render_texture_options){
+			.texture = sample->cat_texture,
+			.dst_box = { .x = x, .y = y },
+		});
 	}
 
-	wlr_renderer_end(sample->renderer);
-	wlr_output_commit(wlr_output);
+	wlr_render_pass_submit(pass);
+	wlr_output_commit_state(wlr_output, &output_state);
+	wlr_output_state_finish(&output_state);
 	sample->last_frame = now;
 }
 
 static void touch_down_notify(struct wl_listener *listener, void *data) {
-	struct wlr_touch_motion_event *event = data;
+	struct wlr_touch_down_event *event = data;
 	struct touch_state *tstate = wl_container_of(listener, tstate, down);
 	struct sample_state *sample = tstate->sample;
-	struct touch_point *point = calloc(1, sizeof(struct touch_point));
+	struct touch_point *point = calloc(1, sizeof(*point));
 	point->touch_id = event->touch_id;
 	point->x = event->x;
 	point->y = event->y;
@@ -130,6 +138,19 @@ static void touch_motion_notify(struct wl_listener *listener, void *data) {
 	}
 }
 
+static void touch_cancel_notify(struct wl_listener *listener, void *data) {
+	struct wlr_touch_cancel_event *event = data;
+	struct touch_state *tstate = wl_container_of(listener, tstate, cancel);
+	struct sample_state *sample = tstate->sample;
+	struct touch_point *point, *tmp;
+	wl_list_for_each_safe(point, tmp, &sample->touch_points, link) {
+		if (point->touch_id == event->touch_id) {
+			wl_list_remove(&point->link);
+			break;
+		}
+	}
+}
+
 static void touch_destroy_notify(struct wl_listener *listener, void *data) {
 	struct touch_state *tstate = wl_container_of(listener, tstate, destroy);
 	wl_list_remove(&tstate->link);
@@ -137,6 +158,7 @@ static void touch_destroy_notify(struct wl_listener *listener, void *data) {
 	wl_list_remove(&tstate->down.link);
 	wl_list_remove(&tstate->up.link);
 	wl_list_remove(&tstate->motion.link);
+	wl_list_remove(&tstate->cancel.link);
 	free(tstate);
 }
 
@@ -153,7 +175,7 @@ static void new_output_notify(struct wl_listener *listener, void *data) {
 
 	wlr_output_init_render(output, sample->allocator, sample->renderer);
 
-	struct sample_output *sample_output = calloc(1, sizeof(struct sample_output));
+	struct sample_output *sample_output = calloc(1, sizeof(*sample_output));
 	sample_output->output = output;
 	sample_output->sample = sample;
 	wl_signal_add(&output->events.frame, &sample_output->frame);
@@ -161,12 +183,15 @@ static void new_output_notify(struct wl_listener *listener, void *data) {
 	wl_signal_add(&output->events.destroy, &sample_output->destroy);
 	sample_output->destroy.notify = output_remove_notify;
 
+	struct wlr_output_state state;
+	wlr_output_state_init(&state);
+	wlr_output_state_set_enabled(&state, true);
 	struct wlr_output_mode *mode = wlr_output_preferred_mode(output);
 	if (mode != NULL) {
-		wlr_output_set_mode(output, mode);
+		wlr_output_state_set_mode(&state, mode);
 	}
-
-	wlr_output_commit(output);
+	wlr_output_commit_state(output, &state);
+	wlr_output_state_finish(&state);
 }
 
 static void keyboard_key_notify(struct wl_listener *listener, void *data) {
@@ -197,7 +222,7 @@ static void new_input_notify(struct wl_listener *listener, void *data) {
 	struct sample_state *sample = wl_container_of(listener, sample, new_input);
 	switch (device->type) {
 	case WLR_INPUT_DEVICE_KEYBOARD:;
-		struct sample_keyboard *keyboard = calloc(1, sizeof(struct sample_keyboard));
+		struct sample_keyboard *keyboard = calloc(1, sizeof(*keyboard));
 		keyboard->wlr_keyboard = wlr_keyboard_from_input_device(device);
 		keyboard->sample = sample;
 		wl_signal_add(&device->events.destroy, &keyboard->destroy);
@@ -220,7 +245,7 @@ static void new_input_notify(struct wl_listener *listener, void *data) {
 		xkb_context_unref(context);
 		break;
 	case WLR_INPUT_DEVICE_TOUCH:;
-		struct touch_state *tstate = calloc(sizeof(struct touch_state), 1);
+		struct touch_state *tstate = calloc(1, sizeof(*tstate));
 		tstate->wlr_touch = wlr_touch_from_input_device(device);
 		tstate->sample = sample;
 		tstate->destroy.notify = touch_destroy_notify;
@@ -231,6 +256,8 @@ static void new_input_notify(struct wl_listener *listener, void *data) {
 		wl_signal_add(&tstate->wlr_touch->events.motion, &tstate->motion);
 		tstate->up.notify = touch_up_notify;
 		wl_signal_add(&tstate->wlr_touch->events.up, &tstate->up);
+		tstate->cancel.notify = touch_cancel_notify;
+		wl_signal_add(&tstate->wlr_touch->events.cancel, &tstate->cancel);
 		wl_list_insert(&sample->touch, &tstate->link);
 		break;
 	default:
@@ -248,7 +275,7 @@ int main(int argc, char *argv[]) {
 	wl_list_init(&state.touch_points);
 	wl_list_init(&state.touch);
 
-	struct wlr_backend *wlr = wlr_backend_autocreate(display);
+	struct wlr_backend *wlr = wlr_backend_autocreate(display, NULL);
 	if (!wlr) {
 		exit(1);
 	}
