@@ -4,11 +4,9 @@
 #include <wayland-util.h>
 #include <wlr/backend/interface.h>
 #include <wlr/backend/multi.h>
-#include <wlr/backend/session.h>
 
 struct wlr_multi_backend {
 	struct wlr_backend backend;
-	struct wlr_session *session;
 
 	struct wl_list backends;
 
