@@ -14,11 +14,10 @@
 
 struct wlr_fullscreen_shell_v1 {
 	struct wl_global *global;
-	struct wl_list resources;
 
 	struct {
 		struct wl_signal destroy;
-		// wlr_fullscreen_shell_v1_present_surface_event
+		// struct wlr_fullscreen_shell_v1_present_surface_event
 		struct wl_signal present_surface;
 	} events;
 
@@ -36,6 +35,5 @@ struct wlr_fullscreen_shell_v1_present_surface_event {
 
 struct wlr_fullscreen_shell_v1 *wlr_fullscreen_shell_v1_create(
 	struct wl_display *display);
-void wlr_fullscreen_shell_v1_destroy(struct wlr_fullscreen_shell_v1 *shell);
 
 #endif
