@@ -17,7 +17,7 @@ struct wlr_presentation_surface_state {
 struct wlr_presentation_surface {
 	struct wlr_presentation_surface_state current, pending;
 
-	struct wlr_addon addon; // wlr_surface::addons
+	struct wlr_addon addon; // wlr_surface.addons
 
 	struct wl_listener surface_commit;
 };
@@ -28,7 +28,7 @@ static void feedback_handle_resource_destroy(struct wl_resource *resource) {
 
 static void feedback_resource_send_presented(
 		struct wl_resource *feedback_resource,
-		struct wlr_presentation_event *event) {
+		const struct wlr_presentation_event *event) {
 	struct wl_client *client = wl_resource_get_client(feedback_resource);
 	struct wl_resource *output_resource;
 	wl_resource_for_each(output_resource, &event->output->resources) {
@@ -61,6 +61,8 @@ static void presentation_surface_addon_destroy(struct wlr_addon *addon) {
 	struct wlr_presentation_surface *p_surface =
 		wl_container_of(addon, p_surface, addon);
 
+	wlr_addon_finish(addon);
+
 	wlr_presentation_feedback_destroy(p_surface->current.feedback);
 	wlr_presentation_feedback_destroy(p_surface->pending.feedback);
 
@@ -120,7 +122,7 @@ static void presentation_handle_feedback(struct wl_client *client,
 
 	struct wlr_presentation_feedback *feedback = p_surface->pending.feedback;
 	if (feedback == NULL) {
-		feedback = calloc(1, sizeof(struct wlr_presentation_feedback));
+		feedback = calloc(1, sizeof(*feedback));
 		if (feedback == NULL) {
 			wl_client_post_no_memory(client);
 			return;
@@ -166,7 +168,7 @@ static void presentation_bind(struct wl_client *client, void *data,
 	wl_resource_set_implementation(resource, &presentation_impl, presentation,
 		NULL);
 
-	wp_presentation_send_clock_id(resource, (uint32_t)presentation->clock);
+	wp_presentation_send_clock_id(resource, CLOCK_MONOTONIC);
 }
 
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
@@ -180,8 +182,7 @@ static void handle_display_destroy(struct wl_listener *listener, void *data) {
 
 struct wlr_presentation *wlr_presentation_create(struct wl_display *display,
 		struct wlr_backend *backend) {
-	struct wlr_presentation *presentation =
-		calloc(1, sizeof(struct wlr_presentation));
+	struct wlr_presentation *presentation = calloc(1, sizeof(*presentation));
 	if (presentation == NULL) {
 		return NULL;
 	}
@@ -193,8 +194,6 @@ struct wlr_presentation *wlr_presentation_create(struct wl_display *display,
 		return NULL;
 	}
 
-	presentation->clock = wlr_backend_get_presentation_clock(backend);
-
 	wl_signal_init(&presentation->events.destroy);
 
 	presentation->display_destroy.notify = handle_display_destroy;
@@ -205,7 +204,7 @@ struct wlr_presentation *wlr_presentation_create(struct wl_display *display,
 
 void wlr_presentation_feedback_send_presented(
 		struct wlr_presentation_feedback *feedback,
-		struct wlr_presentation_event *event) {
+		const struct wlr_presentation_event *event) {
 	struct wl_resource *resource, *tmp;
 	wl_resource_for_each_safe(resource, tmp, &feedback->resources) {
 		feedback_resource_send_presented(resource, event);
@@ -247,13 +246,14 @@ void wlr_presentation_feedback_destroy(
 
 void wlr_presentation_event_from_output(struct wlr_presentation_event *event,
 		const struct wlr_output_event_present *output_event) {
-	memset(event, 0, sizeof(*event));
-	event->output = output_event->output;
-	event->tv_sec = (uint64_t)output_event->when->tv_sec;
-	event->tv_nsec = (uint32_t)output_event->when->tv_nsec;
-	event->refresh = (uint32_t)output_event->refresh;
-	event->seq = (uint64_t)output_event->seq;
-	event->flags = output_event->flags;
+	*event = (struct wlr_presentation_event){
+		.output = output_event->output,
+		.tv_sec = (uint64_t)output_event->when->tv_sec,
+		.tv_nsec = (uint32_t)output_event->when->tv_nsec,
+		.refresh = (uint32_t)output_event->refresh,
+		.seq = (uint64_t)output_event->seq,
+		.flags = output_event->flags,
+	};
 }
 
 static void feedback_unset_output(struct wlr_presentation_feedback *feedback) {
@@ -292,6 +292,9 @@ static void feedback_handle_output_present(struct wl_listener *listener,
 	if (output_event->presented) {
 		struct wlr_presentation_event event = {0};
 		wlr_presentation_event_from_output(&event, output_event);
+		if (!feedback->zero_copy) {
+			event.flags &= ~WP_PRESENTATION_FEEDBACK_KIND_ZERO_COPY;
+		}
 		wlr_presentation_feedback_send_presented(feedback, &event);
 	}
 	wlr_presentation_feedback_destroy(feedback);
@@ -304,9 +307,9 @@ static void feedback_handle_output_destroy(struct wl_listener *listener,
 	wlr_presentation_feedback_destroy(feedback);
 }
 
-void wlr_presentation_surface_sampled_on_output(
+static void presentation_surface_queued_on_output(
 		struct wlr_presentation *presentation, struct wlr_surface *surface,
-		struct wlr_output *output) {
+		struct wlr_output *output, bool zero_copy) {
 	struct wlr_presentation_feedback *feedback =
 		wlr_presentation_surface_sampled(presentation, surface);
 	if (feedback == NULL) {
@@ -315,6 +318,7 @@ void wlr_presentation_surface_sampled_on_output(
 
 	assert(feedback->output == NULL);
 	feedback->output = output;
+	feedback->zero_copy = zero_copy;
 
 	feedback->output_commit.notify = feedback_handle_output_commit;
 	wl_signal_add(&output->events.commit, &feedback->output_commit);
@@ -323,3 +327,17 @@ void wlr_presentation_surface_sampled_on_output(
 	feedback->output_destroy.notify = feedback_handle_output_destroy;
 	wl_signal_add(&output->events.destroy, &feedback->output_destroy);
 }
+
+void wlr_presentation_surface_textured_on_output(
+		struct wlr_presentation *presentation, struct wlr_surface *surface,
+		struct wlr_output *output) {
+	return presentation_surface_queued_on_output(presentation, surface,
+		output, false);
+}
+
+void wlr_presentation_surface_scanned_out_on_output(
+		struct wlr_presentation *presentation, struct wlr_surface *surface,
+		struct wlr_output *output) {
+	return presentation_surface_queued_on_output(presentation, surface,
+		output, true);
+}
