@@ -15,10 +15,11 @@ struct wlr_tablet *wlr_tablet_from_input_device(
 
 void wlr_tablet_init(struct wlr_tablet *tablet,
 		const struct wlr_tablet_impl *impl, const char *name) {
-	memset(tablet, 0, sizeof(*tablet));
+	*tablet = (struct wlr_tablet){
+		.impl = impl,
+	};
 	wlr_input_device_init(&tablet->base, WLR_INPUT_DEVICE_TABLET_TOOL, name);
 
-	tablet->impl = impl;
 	wl_signal_init(&tablet->events.axis);
 	wl_signal_init(&tablet->events.proximity);
 	wl_signal_init(&tablet->events.tip);
