@@ -323,13 +323,8 @@ static void xdg_popup_handle_reposition(
 		return;
 	}
 
-	struct wlr_xdg_positioner *positioner = wlr_xdg_positioner_from_resource(positioner_resource);
-	if (!wlr_xdg_positioner_is_complete(positioner)) {
-		wl_resource_post_error(popup->base->client->resource,
-			XDG_WM_BASE_ERROR_INVALID_POSITIONER, "positioner object is not complete");
-		return;
-	}
-
+	struct wlr_xdg_positioner *positioner =
+		wlr_xdg_positioner_from_resource(positioner_resource);
 	wlr_xdg_positioner_rules_get_geometry(
 		&positioner->rules, &popup->scheduled.geometry);
 	popup->scheduled.rules = positioner->rules;
@@ -372,11 +367,14 @@ static void xdg_popup_handle_resource_destroy(struct wl_resource *resource) {
 	wlr_xdg_popup_destroy(popup);
 }
 
-void create_xdg_popup(struct wlr_xdg_surface *surface, struct wlr_xdg_surface *parent,
+void create_xdg_popup(struct wlr_xdg_surface *surface,
+		struct wlr_xdg_surface *parent,
 		struct wlr_xdg_positioner *positioner, uint32_t id) {
-	if (!wlr_xdg_positioner_is_complete(positioner)) {
+	if (positioner->rules.size.width == 0 ||
+			positioner->rules.anchor_rect.width == 0) {
 		wl_resource_post_error(surface->client->resource,
-			XDG_WM_BASE_ERROR_INVALID_POSITIONER, "positioner object is not complete");
+			XDG_WM_BASE_ERROR_INVALID_POSITIONER,
+			"positioner object is not complete");
 		return;
 	}
 
@@ -422,15 +420,12 @@ void create_xdg_popup(struct wlr_xdg_surface *surface, struct wlr_xdg_surface *p
 	if (parent) {
 		surface->popup->parent = parent->surface;
 		wl_list_insert(&parent->popups, &surface->popup->link);
+		wl_signal_emit_mutable(&parent->events.new_popup, surface->popup);
 	} else {
 		wl_list_init(&surface->popup->link);
 	}
 
 	set_xdg_surface_role_object(surface, surface->popup->resource);
-
-	if (parent) {
-		wl_signal_emit_mutable(&parent->events.new_popup, surface->popup);
-	}
 }
 
 void reset_xdg_popup(struct wlr_xdg_popup *popup) {
