@@ -133,34 +133,6 @@ static int xwayland_surface_handle_ping_timeout(void *data) {
 	return 1;
 }
 
-static void read_surface_client_id(struct wlr_xwm *xwm,
-		struct wlr_xwayland_surface *xsurface,
-		xcb_res_query_client_ids_cookie_t cookie) {
-	xcb_res_query_client_ids_reply_t *reply = xcb_res_query_client_ids_reply(
-		xwm->xcb_conn, cookie, NULL);
-	if (reply == NULL) {
-		return;
-	}
-
-	uint32_t *pid = NULL;
-	xcb_res_client_id_value_iterator_t iter =
-		xcb_res_query_client_ids_ids_iterator(reply);
-	while (iter.rem > 0) {
-		if (iter.data->spec.mask & XCB_RES_CLIENT_ID_MASK_LOCAL_CLIENT_PID &&
-				xcb_res_client_id_value_value_length(iter.data) > 0) {
-			pid = xcb_res_client_id_value_value(iter.data);
-			break;
-		}
-		xcb_res_client_id_value_next(&iter);
-	}
-	if (pid == NULL) {
-		free(reply);
-		return;
-	}
-	xsurface->pid = *pid;
-	free(reply);
-}
-
 static struct wlr_xwayland_surface *xwayland_surface_create(
 		struct wlr_xwm *xwm, xcb_window_t window_id, int16_t x, int16_t y,
 		uint16_t width, uint16_t height, bool override_redirect) {
@@ -173,15 +145,6 @@ static struct wlr_xwayland_surface *xwayland_surface_create(
 	xcb_get_geometry_cookie_t geometry_cookie =
 		xcb_get_geometry(xwm->xcb_conn, window_id);
 
-	xcb_res_query_client_ids_cookie_t client_id_cookie;
-	if (xwm->xres) {
-		xcb_res_client_id_spec_t spec = {
-			.client = window_id,
-			.mask = XCB_RES_CLIENT_ID_MASK_LOCAL_CLIENT_PID
-		};
-		client_id_cookie = xcb_res_query_client_ids(xwm->xcb_conn, 1, &spec);
-	}
-
 	uint32_t values[1];
 	values[0] =
 		XCB_EVENT_MASK_FOCUS_CHANGE |
@@ -242,10 +205,6 @@ static struct wlr_xwayland_surface *xwayland_surface_create(
 
 	wl_list_insert(&xwm->surfaces, &surface->link);
 
-	if (xwm->xres) {
-		read_surface_client_id(xwm, surface, client_id_cookie);
-	}
-
 	wl_signal_emit_mutable(&xwm->xwayland->events.new_surface, surface);
 
 	return surface;
@@ -258,25 +217,6 @@ static void xwm_set_net_active_window(struct wlr_xwm *xwm,
 			xwm->atoms[WINDOW], 32, 1, &window);
 }
 
-/*
- * Wrapper for xcb_send_event, which ensures that the event data is 32 byte big.
- */
-xcb_void_cookie_t xwm_send_event_with_size(xcb_connection_t *c,
-		uint8_t propagate, xcb_window_t destination,
-		uint32_t event_mask, const void *event, uint32_t length)
-{
-	if (length == 32) {
-		return xcb_send_event(c, propagate, destination, event_mask, event);
-	} else if (length < 32) {
-		char buf[32];
-		memcpy(buf, event, length);
-		memset(buf + length, 0, 32 - length);
-		return xcb_send_event(c, propagate, destination, event_mask, buf);
-	} else {
-		assert(false && "Event too long");
-	}
-}
-
 static void xwm_send_wm_message(struct wlr_xwayland_surface *surface,
 		xcb_client_message_data_t *data, uint32_t event_mask) {
 	struct wlr_xwm *xwm = surface->xwm;
@@ -290,12 +230,11 @@ static void xwm_send_wm_message(struct wlr_xwayland_surface *surface,
 		.data = *data,
 	};
 
-	xwm_send_event_with_size(xwm->xcb_conn,
+	xcb_send_event(xwm->xcb_conn,
 		0, // propagate
 		surface->window_id,
 		event_mask,
-		&event,
-		sizeof(event));
+		(const char *)&event);
 	xcb_flush(xwm->xcb_conn);
 }
 
@@ -656,6 +595,34 @@ static void read_surface_parent(struct wlr_xwm *xwm,
 	wl_signal_emit_mutable(&xsurface->events.set_parent, NULL);
 }
 
+static void read_surface_client_id(struct wlr_xwm *xwm,
+		struct wlr_xwayland_surface *xsurface,
+		xcb_res_query_client_ids_cookie_t cookie) {
+	xcb_res_query_client_ids_reply_t *reply = xcb_res_query_client_ids_reply(
+		xwm->xcb_conn, cookie,  NULL);
+	if (reply == NULL) {
+		return;
+	}
+
+	uint32_t *pid = NULL;
+	xcb_res_client_id_value_iterator_t iter =
+		xcb_res_query_client_ids_ids_iterator(reply);
+	while (iter.rem > 0) {
+		if (iter.data->spec.mask & XCB_RES_CLIENT_ID_MASK_LOCAL_CLIENT_PID &&
+				xcb_res_client_id_value_value_length(iter.data) > 0) {
+			pid = xcb_res_client_id_value_value(iter.data);
+			break;
+		}
+		xcb_res_client_id_value_next(&iter);
+	}
+	if (pid == NULL) {
+		free(reply);
+		return;
+	}
+	xsurface->pid = *pid;
+	free(reply);
+}
+
 static void read_surface_window_type(struct wlr_xwm *xwm,
 		struct wlr_xwayland_surface *xsurface,
 		xcb_get_property_reply_t *reply) {
@@ -954,17 +921,30 @@ static void xwayland_surface_associate(struct wlr_xwm *xwm,
 			props[i], XCB_ATOM_ANY, 0, 2048);
 	}
 
+	xcb_res_query_client_ids_cookie_t client_id_cookie;
+	if (xwm->xres) {
+		xcb_res_client_id_spec_t spec = {
+			.client = xsurface->window_id,
+			.mask = XCB_RES_CLIENT_ID_MASK_LOCAL_CLIENT_PID
+		};
+		client_id_cookie = xcb_res_query_client_ids(xwm->xcb_conn, 1, &spec);
+	}
+
 	for (size_t i = 0; i < sizeof(props) / sizeof(props[0]); i++) {
 		xcb_get_property_reply_t *reply =
 			xcb_get_property_reply(xwm->xcb_conn, cookies[i], NULL);
 		if (reply == NULL) {
 			wlr_log(WLR_ERROR, "Failed to get window property");
-			continue;
+			return;
 		}
 		read_surface_property(xwm, xsurface, props[i], reply);
 		free(reply);
 	}
 
+	if (xwm->xres) {
+		read_surface_client_id(xwm, xsurface, client_id_cookie);
+	}
+
 	wl_signal_emit_mutable(&xsurface->events.associate, NULL);
 }
 
@@ -1815,10 +1795,9 @@ void wlr_xwayland_surface_configure(struct wlr_xwayland_surface *xsurface,
 			.height = height,
 		};
 
-		xwm_send_event_with_size(xwm->xcb_conn, 0, xsurface->window_id,
+		xcb_send_event(xwm->xcb_conn, 0, xsurface->window_id,
 			XCB_EVENT_MASK_STRUCTURE_NOTIFY,
-			&configure_notify,
-			sizeof(configure_notify));
+			(const char *)&configure_notify);
 	}
 
 	xcb_flush(xwm->xcb_conn);
